#!/usr/bin/env python
# -*- coding: utf-8 -*-

import struct
import time     # sleep
from classes import df_Common
from classes import df_SFC

# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 0

#---------------------------------------------------------------------------
# SFメモリカセットへのコマンド送信
#       戻り値：
#               retCode: 
#                    0:成功
#                   -1:失敗
#
#       メモ：
#---------------------------------------------------------------------------
def send_SFM(s, dataByte):
    df_SFC.writeBytes_SFC(s, 0x2400, bytes([0x09]))
    
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 1)

    df_SFC.writeBytes_SFC(s, 0x2401, bytes([0x28]))
    df_SFC.writeBytes_SFC(s, 0x2401, bytes([0x84]))

    if (retCode == 0 and retAry[0] == 0x7d):
        df_SFC.writeBytes_SFC(s, 0x2400, bytes([0x06]))
        df_SFC.writeBytes_SFC(s, 0x2400, bytes([0x39]))
    
    df_SFC.writeBytes_SFC(s, 0x2400, bytes([dataByte]))  # command = 0x04
    #time.sleep(0.2)

    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 1)

#---------------------------------------------------------------------------
# [SFC] SFメモリカセットWakeup(HIROM:ALL)状態にする
#       戻り値：
#            0:成功
#           -1:失敗
#       メモ：
#---------------------------------------------------------------------------
def wakeup_SFM(s):
    # 1-1. 2Fコマンド送信
    df_Common.putLog('2Fコマンド送信')
    s.write(bytes([
        0x2f, 0x00, 0x00, 0x00,             # 
        0x08, 0x00, 0x00, 0x00,             # 不明 8
        0x00, 0x00, 0x00, 0x00,             # 不明 0
        0x00, 0x00, 0x00, 0x00,             # 不明 0
        0x01, 0x00, 0x00, 0x00              # 不明 1
        ]))

    # 1-2. 2Fコマンド応答待ち
    respWaitCount = 0
    while True:
        if s.in_waiting > 0:
            df_Common.putLog('受信データあり: %d' % s.in_waiting)
            break
        else:
            respWaitCount = respWaitCount + 1
            df_Common.putLog('0.01秒待機。')
            time.sleep(0.01)
            if(respWaitCount > 40):
                print('wakeup_SFM() ERR: 2Fコマンド応答タイムアウト')
                return -1

    # 1-3. 2Fコマンド応答の受信と判定
    readStr = s.read(8)
    readBytes = bytes(readStr)
    df_Common.putLog('2Fコマンド応答受信。')
    df_Common.debugDump(readBytes)

    if readBytes == bytes([0,0,0,0,0,0,0,0]):
        df_Common.putLog('肯定応答を受信。')
    else:
        #df_Common.putLog('肯定応答でない応答を受信。')
        #print('ERR: 2Fコマンドに肯定応答でない応答')
        #return -2, retAry
        df_Common.putLog('肯定応答でない応答を受信。続行。')

    time.sleep(0.5)

    enableCount = 0
    retryCount  = 10
    waitSec = 0.1
    while True:
        df_Common.putLog('SFメモリWakeupコマンド(HiMEM All)送信')
        send_SFM(s, 0x04)
        
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 1)

        if (retCode == 0):
            df_Common.putLog('Wakeupコマンド完了後の0x2400の値: 0x%02x' % retAry[0])
            if retAry[0] == 0x2a:
                df_Common.putLog('Wakeupコマンド正常終了')
                break
            else:
                time.sleep(waitSec)
                enableCount = enableCount + 1
                if enableCount > retryCount:
                    print('wakeup_SFM() Err: リトライ上限到達。Wakeupコマンド失敗。(値不正)')
                    return -1
                else:
                    df_Common.putLog('Err: Wakeupコマンド完了後の0x2400の値!=0x2a。0.1秒後リトライ実施。')
                    time.sleep(waitSec)
                    continue
        else:
            enableCount = enableCount + 1
            if enableCount > retryCount:
                print('wakeup_SFM() Err: リトライ上限到達。Wakeupコマンド失敗。(値取得失敗)')
                return -1
            else:
                df_Common.putLog('Err: Wakeupコマンド完了後の0x2400の値の取得に失敗。0.1秒後リトライ実施。')
                time.sleep(waitSec)
                continue

    # ここまでWakeup処理
    return 0

#---------------------------------------------------------------------------
# [SFC] SFメモリカセットの書き換え サブルーチン
#           idFlash_SFM
#       メモ：
#---------------------------------------------------------------------------
def idFlash_SFM(s, startBank):
    # HiROMとみなす
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xAA]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x90]))
    
    # Get Maker ID
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, (startBank << 16) + 0x0000, 1)
    if retCode != 0:
        print('Err: Maker ID読み取り処理失敗。')
        return 0, 0
    else:
        makerID = retAry[0]

    # Get Device ID
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, (startBank << 16) + 0x0002, 1)
    if retCode != 0:
        print('Err: Device ID読み取り処理失敗。')
        return 0, 0
    else:
        deviceID = retAry[0]

    return makerID, deviceID

#---------------------------------------------------------------------------
# [SFC][SFM] SFメモリカセットの書き換え サブルーチン
#           resetFlash_SFM
#       メモ：
#---------------------------------------------------------------------------
def resetFlash_SFM(s, startBank):
    # HiROMとみなす
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xAA]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xF0]))


#---------------------------------------------------------------------------
# [SFC][SFM] マップ情報表示
#
#       メモ：
#---------------------------------------------------------------------------
def printMapping_SFM(s):
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x38]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0xd0]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x71]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x72]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x75]))

    
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, (0xc0 << 16) + 0xff00, 0x50)
    if retCode != 0:
        print('printMapping_SFM(): Err: Flash %02xの状態読み込み失敗。' % 0xc0)
        return
    df_Common.dump(retAry, 0xc0ff00)

    # reset flash
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0xAAAA, bytes([0xf0]))

    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0xAAAA, bytes([0xf0]))

#---------------------------------------------------------------------------
# [SFC] マップ情報読み込み
#
#       メモ：
#---------------------------------------------------------------------------
def readMapping_SFM(s):
    retAry = []
    
    # c0隠しセクターへのアクセス有効化
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x38]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0xd0]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x71]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x72]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x75]))
    
    # 配列に読み込み
    (retCode, readBuf) = df_Common.readSizeByType(s, 0x07, (0xc0 << 16) + 0xff00, 256)
    if retCode != 0:
        print('readMapping_SFM() Err: 隠しセクター0xc0ff00読み取り処理失敗。')
        return -1, retAry

    # 返却配列に追加
    retAry.extend(readBuf)
    
    # e0隠しセクターへのアクセス有効化
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x38]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0xd0]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x71]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x72]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x75]))
    
    # 配列に読み込み
    (retCode, readBuf) = df_Common.readSizeByType(s, 0x07, (0xe0 << 16) + 0xff00, 256)
    if retCode != 0:
        print('readMapping_SFM() Err: 隠しセクター0xe0ff00読み取り処理失敗。')
        return -1, retAry

    # 返却配列に追加
    retAry.extend(readBuf)

    # reset flash
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0xAAAA, bytes([0xf0]))

    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0xAAAA, bytes([0xf0]))

    return 0, retAry

#---------------------------------------------------------------------------
# [SFC][SFM] セクタープロテクト情報表示
#
#       メモ：
#---------------------------------------------------------------------------
def readSectorProtection_SFM(s, startBank):
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x90]))

    print('[コマンド90の結果]')

    print('Sector MID  DID  Protect')
    i = 0
    while i <= 0x1f:
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, (startBank + i)<<16, 5)
        if retCode != 0:
            print('Err: readSectorProtection_SFM(): readSizeByType()失敗。')
            return

        print('0x%02x   0x%02x 0x%02x 0x%02x' % (startBank + i, retAry[0], retAry[2], retAry[4]), end='')
        if retAry[4] == 0xc2:
            print('   **** Protected Sector ****')
        else:
            print('')
        i = i + 1

    print('[セクター毎のステータスレジスタ]')
    i = 0
    while i <= 0x1f:
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x70]))
        
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, (startBank + i)<<16, 1)
        if retCode != 0:
            print('Err: readSectorProtection_SFM(): readSizeByType()失敗。')
            return

        print('Sector:0x%02x SRD:0x%02x' % (startBank + i, retAry[0]))
        i = i + 1

#---------------------------------------------------------------------------
# [SFM] マップ情報イレーズコマンド送信＆コマンド終了まで待機
#       メモ：
#---------------------------------------------------------------------------
def sendHiddenSectorEraseCommand_SFM(s, startBank):
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x77]))

    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xe0]))
    
    while True:
        time.sleep(0.1)
        (retCode, readBuf) = df_Common.readSizeByType(s, 0x07, (startBank << 16) + 0, 1)
        if retCode == -1:
            print('Err: sendHiddenSectorEraseCommand_SFM() Flash %02x 読み取り失敗。' % startBank)
            return

        # 以下の処理における、Flash 0xc0を対象にした時の挙動
        # 
        # (1)「if (readBuf[0] & 0x80) != 0:」を有効にした場合
        #       0xc00000==0xe0が返ってきて、このwhileループを抜ける
        #
        # (2)「if readBuf[0] == 0x80:」を有効にした場合
        #       最初のみ0xc00000==0xe0となり、その後の読み出しでは
        #       0xc00000==0xe0が続く。
        #       結果、このwhileループを抜けられない
        #
        
        if (readBuf[0] & 0x80) != 0:
        #if readBuf[0] == 0x80:
            print('sendHiddenSectorEraseCommand_SFM() %02x0000 bit7==Onを確認(%02x)' % (startBank, readBuf[0]))
            break
        print('sendHiddenSectorEraseCommand_SFM() %02x0000 bit7==On待機(%02x)' % (startBank, readBuf[0]))

#---------------------------------------------------------------------------
# [SFM] セクタープロテクトを有効化する
#       メモ：
#---------------------------------------------------------------------------
def setSectorProtect_SFM(s, startBank):
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x60]))

    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x20]))

    while True:
        time.sleep(0.1)
        (retCode, readBuf) = df_Common.readSizeByType(s, 0x07, (startBank << 16) + 0, 1)
        if retCode == -1:
            print('Err: setSectorProtect_SFM() Flash %02x 読み取り失敗。' % startBank)
            return

        if (readBuf[0] & 0x80) != 0:
            print('setSectorProtect_SFM() %02x0000 bit7==Onを確認(%02x)' % (startBank, readBuf[0]))
            break
        print('setSectorProtect_SFM() %02x0000 bit7==On待機(%02x)' % (startBank, readBuf[0]))

#---------------------------------------------------------------------------
# [SFM] Flashプロテクトを有効化する
#       メモ：
#---------------------------------------------------------------------------
def setFlashProtect_SFM(s, startBank):
    wakeup_SFM(s)
    
    print('0.3秒待機します。')
    time.sleep(0.3)

    # Wakeup状態チェック
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 16)
    if retCode != 0:
        print('Err: setFlashProtect_SFM() Wakeup状態読み取り処理失敗。')
        return
    else:
        if retAry[0] == 0x2a:
            print('setFlashProtect_SFM() SFメモリカセットのWakeup処理に成功しました。')
        else:
            print('Err: setFlashProtect_SFM() SFメモリカセットのWakeup処理に失敗しました。')
            return
            
    # ここまで来たら、Wakeup(HIMEM:ALL)に成功している

    send_SFM(s, 0x03)

    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2401, 1)
    if (retCode == 0):
        print('setFlashProtect_SFM(): プロテクト有効化コマンド完了後の0x2401の値: 0x%02x' % retAry[0])
        if retAry[0] == 0x00:
            print('プロテクト有効化成功。')
        else:
            print('Err: プロテクト有効化コマンド送信後の状態が不正。')
            return False
    else:
        print('Err: プロテクト有効化コマンド送信後の状態取得に失敗。')
        return False

#---------------------------------------------------------------------------
# [SFM] Flashプロテクトを無効化する
#       メモ：
#---------------------------------------------------------------------------
def unsetFlashProtect_SFM(s, startBank):
    wakeup_SFM(s)
    
    print('0.3秒待機します。')
    time.sleep(0.3)

    # Wakeup状態チェック
    if isWakeup_SFM(s) != True:
        print('Err: SFメモリカセットがWakeup状態ではありません。')
        return

    # ここまで来たら、Wakeup(HIMEM:ALL)に成功している

    send_SFM(s, 0x02)

    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2401, 1)
    if (retCode == 0):
        print('unsetFlashProtect_SFM(): プロテクト有効化コマンド完了後の0x2401の値: 0x%02x' % retAry[0])
        if retAry[0] == 0x04:
            print('プロテクト無効化成功。')
        else:
            print('Err: プロテクト無効化コマンド送信後の状態が不正。')
            return False
    else:
        print('Err: プロテクト無効化コマンド送信後の状態取得に失敗。')
        return False


#---------------------------------------------------------------------------
# [SFM] セクタープロテクトを無効化する
#       メモ：
#---------------------------------------------------------------------------
def unsetSectorProtect_SFM(s, startBank):
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x60]))

    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x40]))

    while True:
        time.sleep(0.1)
        (retCode, readBuf) = df_Common.readSizeByType(s, 0x07, (startBank << 16) + 0, 1)
        if retCode == -1:
            print('Err: unsetSectorProtect_SFM() Flash %02x 読み取り失敗。' % startBank)
            return

        if (readBuf[0] & 0x80) != 0:
            print('unsetSectorProtect_SFM() %02x0000 bit7==Onを確認(%02x)' % (startBank, readBuf[0]))
            break
        print('unsetSectorProtect_SFM() %02x0000 bit7==On待機(%02x)' % (startBank, readBuf[0]))

#---------------------------------------------------------------------------
# [SFM] SFメモリカセットはWakeup状態か？
#       メモ：
#---------------------------------------------------------------------------
def isWakeup_SFM(s):
    # Wakeup状態チェック
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 16)
    if retCode != 0:
        print('Err: isWakeup_SFM() カセット読み取り処理失敗。')
        return False
    else:
        if retAry[0] == 0x2a:
            #print('isWakeup_SFM() SFメモリカセットはWakeup中です。')
            return True
        else:
            print('isWakeup_SFM() SFメモリカセットはWakeup中ではありません。')
            return False

#---------------------------------------------------------------------------
# [SFC] SFメモリカセットの書き換え サブルーチン
#           blankcheckMapping_SFM
#       メモ：マップ情報領域が全部0xffかどうかを返す
#
#       1: 全部0xff
#       0: 0xffでないデータあり
#---------------------------------------------------------------------------
def blankcheckMapping_SFM(s):
    result = 1
    
    # c0隠しセクターへのアクセス有効化
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x38]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0xd0]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x71]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x72]))
    df_SFC.writeBytes_SFC(s, 0xc00000, bytes([0x75]))
    
    # 配列に読み込み
    (retCode, readBuf) = df_Common.readSizeByType(s, 0x07, (0xc0 << 16) + 0xff00, 256)
    if retCode != 0:
        print('blankcheckMapping_SFM() Err: 隠しセクター0xc0ff00読み取り処理失敗。')
        result =  0

    print('blankcheckMapping_SFM() 0xc0ff00～: %02x %02x %02x %02x' % 
        (readBuf[0],readBuf[1],readBuf[2],readBuf[3]))

    # readBufはすべて0xffか？
    if (readBuf[1:] == readBuf[:-1]) and readBuf[0] == 0xff:
        pass
    else:
        result = 0

    # e0隠しセクターへのアクセス有効化
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x38]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0xd0]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x71]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x72]))
    df_SFC.writeBytes_SFC(s, 0xe00000, bytes([0x75]))
    
    # 配列に読み込み
    (retCode, readBuf) = df_Common.readSizeByType(s, 0x07, (0xe0 << 16) + 0xff00, 256)
    if retCode != 0:
        print('blankcheckMapping_SFM() Err: 隠しセクター0xe0ff00読み取り処理失敗。')
        result = 0

    print('blankcheckMapping_SFM() 0xe0ff00～: %02x %02x %02x %02x' % 
        (readBuf[0],readBuf[1],readBuf[2],readBuf[3]))

    # readBufはすべて0xffか？
    if (readBuf[1:] == readBuf[:-1]) and readBuf[0] == 0xff:
        pass
    else:
        result = 0

    # reset flash
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (0xc0 << 16) + 0xAAAA, bytes([0xf0]))

    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (0xe0 << 16) + 0xAAAA, bytes([0xf0]))

    return result


#---------------------------------------------------------------------------
# [SFC] マップ情報消去
#
#       メモ：
#---------------------------------------------------------------------------
def eraseMapping(s, startBank):
    # unlockHirom()は省略

    print('eraseMapping() ID確認の実行。')
    # ID確認
    (mID, dID) = idFlash_SFM(s, startBank)
    if (mID != 0xc2) or (dID != 0xf3):
        print('eraseMapping() Err: Flash %02x Maker ID/Device IDが不正です。(%02x%02x)' % (startBank, mID, dID))
        return False
    
    # リセット
    print('eraseMapping() flashリセット実行。')
    resetFlash_SFM(s, startBank)
    time.sleep(0.1)

    # 隠しセクター消去
    print('eraseMapping() マップ情報消去実行。')
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x77]))

    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xe0]))

    # 処理完了待ち
    print('eraseMapping() 処理完了待ち実行。')
    if busyCheck_SFM(s, startBank) != True:
        print('Err: eraseMapping() 処理完了待ち失敗。')
        return False
    
    return True
    

#---------------------------------------------------------------------------
# [SFC] SFメモリカセットの書き換え サブルーチン
#           busyCheck_SFM
#       メモ：
#
#---------------------------------------------------------------------------
def busyCheck_SFM(s, startBank):
    waitSec = 0.2
    timeoutCnt = 0
    timeoutMax = 10

    # print('Flash %02xの状態がReadyになるまで待機します。' % startBank)

    while True:
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, (startBank << 16) + 0, 1)
        if retCode != 0:
            print('Err: Flash %02xの状態読み込み失敗。' % startBank)
            timeoutCnt = timeoutCnt + 1
            if timeoutCnt < timeoutMax:
                time.sleep(waitSec)
                continue
            else:
                print('Err: Flash %02xの状態がReadyになりませんでした。(Timeout)' % startBank)
                return False
        else:
            if (retAry[0] & 0x80) != 0:
            #if retAry[0] == 0x80:
                #
                #print('Flash %02xの状態がReadyになりました。(0x%02x)' % (startBank, retAry[0]))
                return True
            else:
                #
                #print('Flash %02xの状態がReadyではありません(0x%02x)' % (startBank, retAry[0]))
                timeoutCnt = timeoutCnt + 1
                if timeoutCnt < timeoutMax:
                    time.sleep(waitSec)
                    continue
                else:
                    print('Err: Flash %02xの状態がReadyになりませんでした。(Timeout)' % startBank)
                    return False

#---------------------------------------------------------------------------
# [SFC][SFM] SFメモリカセットの書き換え サブルーチン
#           unlockHirom
#       メモ：Wakeupの簡略版？
#
#---------------------------------------------------------------------------
def unlockHirom(s):
    # HiMEM Allコマンド送信 & 状態確認
    #print('HiMEM Allコマンド送信')
    send_SFM(s, 0x04)

    # Wakeup状態チェック
    if isWakeup_SFM(s) != True:
        print('unlockHirom() Err: SFメモリカセットがWakeup状態ではありません。')
        return False


    # ライトプロテクト解除コマンド送信 & 状態確認
    if unprotectCart_SFM(s) == False:
        print('Err: unlockHirom() ライトプロテクト解除失敗。')
        return False

    return True


#---------------------------------------------------------------------------
# [SFC][SFM] ライトプロテクト解除
#           SFメモリカセット内のマイコンへのプロテクト解除コマンド送信
#           セクタープロテクトとは別の概念
#---------------------------------------------------------------------------
def unprotectCart_SFM(s):
    # ライトプロテクト解除コマンド送信 & 状態確認
    send_SFM(s, 0x02)

    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2401, 1)
    if (retCode == 0):
        #df_Common.debugDump('ライトプロテクト解除コマンド完了後の0x2401の値: 0x%02x' % retAry[0])
        if retAry[0] != 0x04:
            print('unprotectCart_SFM() Err: ライトプロテクト解除コマンド送信後の状態が不正。')
            return False
    else:
        print('unprotectCart_SFM() Err: ライトプロテクト解除コマンド送信後の状態取得に失敗。')
        return False

    #df_Common.debugDump('unprotectCart_SFM() ライトプロテクト解除成功。')
    return True

#---------------------------------------------------------------------------
# [SFC][SFM] SFメモリカセットの書き換え サブルーチン
#           マップ情報のカセットへの書き込み
# 引数：
#   startBank:      バンク番号
#   mapDataBytes:   マップ情報を読み込んだbytes配列
#   startIdx:       書き込み開始位置（mapDataBytes上の位置）
#---------------------------------------------------------------------------
def writeMapping_SFM(s,startBank, mapDataBytes, startIdx):

    if unlockHirom(s) != True:
        print('writeMapping_SFM() Err: unlockHirom()失敗。')
        return
    
    # ID確認
    (mID, dID) = idFlash_SFM(s, startBank)
    if (mID != 0xc2) or (dID != 0xf3):
        print(
            'writeMapping_SFM() Err: Flash %02x Maker ID/Device IDが不正です。(%02x%02x)' % 
                (startBank, mID, dID))
        return False
    
    # リセット
    #print('writeMapping_SFM() flashリセット実行。')
    resetFlash_SFM(s, startBank)
    
    # 書き換え
    print('Flash %02x 隠しセクターの書き換えを行います。' % startBank)
    addr = 0
    while addr < 0xff:
        print('\r[%.2f%%]アドレス0x%06x書き換え中...' % 
            ((addr / 0xff)*100, (startBank << 16) + addr + 0xff00), end='')
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x77]))

        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xaa]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x99]))

        # 128バイト分
        df_SFC.writeBytes_SFC(s, 
            (startBank << 16) + addr + 0xff00, 
            mapDataBytes[startIdx + addr:startIdx + addr + 128]
        )

        # 最後の1バイトは2回
        df_SFC.writeBytes_SFC(s, 
            (startBank << 16) + addr + 0xff00 + 127, 
            bytes([mapDataBytes[startIdx + addr + 127]])
        )

        # 128(+1)バイト書き込み後、Readyを待つ
        busyCheck_SFM(s, startBank)
        
        # Ready後、次へ
        addr = addr + 128

    print('\r[100%]完了。                                    ')


#---------------------------------------------------------------------------
# [SFC][SFM] SFメモリカセットの書き換え サブルーチン
#           blankcheck_SFM
#       メモ：SFメモリが全64bank(4096KByte=32MBit)だとして、
#             その半分の32バンク分が全部0xffかどうかを返す
#
#       1: 全部0xff
#       0: 0xffでないデータあり
#---------------------------------------------------------------------------
def blankcheck_SFM(s, startBank):
    dumpSize1time = 1024
    addr = 0
    while addr < 2048 * 1024:
        realAddr = (startBank << 16) + addr
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, realAddr, dumpSize1time)
        if retCode == 0:
            # retAryはすべて0xffか？
            if (retAry[1:] == retAry[:-1]) and retAry[0] == 0xff:
                addr = addr + dumpSize1time
                continue
            else:
                return 0
        else:
            print('Err: blankcheck_SFM()読み取り処理失敗。')
            return 0
    return 1

#---------------------------------------------------------------------------
# [SFC] SFメモリカセットの書き換え サブルーチン
#           eraseFlash_SFM
#       メモ：
#
#---------------------------------------------------------------------------
def eraseFlash_SFM(s, startBank):
    # HiROMとみなす
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xAA]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x80]))
    
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xAA]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
    df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0x10]))

    if busyCheck_SFM(s, startBank) == True:
        return True
    else:
        return False

#---------------------------------------------------------------------------
# [SFC] SFメモリカセットの書き換え サブルーチン
#           writeFlash_SFM
#       メモ：
#
#---------------------------------------------------------------------------
def writeFlash_SFM(s, startBank, writeAry, startIdx):
    # 常にHiROMとみなす
    # 本関数では1コのFlash分、16Mbit(0x200000)を書き換える処理とする。

    #print('Flash %02xの書き換えを行います。' % startBank)

    addr = 0
    while addr < 0x200000:
        print('\r[%.2f%%]アドレス0x%06x書き換え中...' % 
            ((addr / 0x200000)*100, (startBank << 16) + addr), end='')
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xAA]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0x5554, bytes([0x55]))
        df_SFC.writeBytes_SFC(s, (startBank << 16) + 0xAAAA, bytes([0xA0]))

        # 128バイト分
        df_SFC.writeBytes_SFC(s, 
            (startBank << 16) + addr, 
            writeAry[startIdx + addr:startIdx + addr + 128])
        # 最後の1バイトは2回
        df_SFC.writeBytes_SFC(s, 
            (startBank << 16) + addr + 127, 
            bytes([writeAry[startIdx + addr + 127]]))
        
        # 128(+1)バイト書き込み後、Readyを待つ
        busyCheck_SFM(s, startBank)

        # Ready後、次へ
        addr = addr + 128

    print('\r[100%]完了。                                    ')


#---------------------------------------------------------------------------
# [SFC][SFM] SFメモリカセットの書き換え サブルーチン
#           カセット上のFlash1個分のbytesを読み取り
#       メモ：
#---------------------------------------------------------------------------
def readFlash_SFM(s, startAddr):
    retBytes = []
    
    # SFメモリwakeup処理
    wakeup_SFM(s)
    
    # Wakeup状態チェック
    if isWakeup_SFM(s) != True:
        print('readFlash_SFM() Err: SFメモリカセットがWakeup状態ではありません。')
        return False, retBytes

    addr = 0
    dumpSize1time = 1024
    allSize = 8 * 512 *1024 /2   # SFメモリカセットの半分

    while addr < allSize:
        # ファイル出力用配列に追加
        realAddr = df_SFC.calcRealAddrSFC_HiROM(addr + startAddr)
        
        #print('\rアドレス {:06x} から読み出し中...'.format(realAddr), end='')
        print('\r[%.2f%%]アドレス {:06x} から読み出し中...'.format(realAddr) % 
            ((addr / allSize)*100), end='')

        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, realAddr, dumpSize1time)
        if retCode == 0:
            retBytes.extend(retAry)
            addr = addr + dumpSize1time
        else:
            print('リトライ実施(retCode: %d)' % retCode)
            #s.rtscts = 1
            continue
        
    print('\r[100%]完了。                                    ')
    return True, retBytes


#---------------------------------------------------------------------------
# [SFC] SFメモリカセットの書き換え
#
#       メモ：
#---------------------------------------------------------------------------
def write_SFM(s, bankNum, byteArrayToWrite, startIdx):
    if unlockHirom(s) != True:
        print('write_SFM() Err: unlockHirom()失敗。')
        return

    (mID, dID) = idFlash_SFM(s, 0xc0)
    if (mID != 0xc2) or (dID != 0xf3):
        print('write_SFM() Err: Flash IDが不正です。(%02x%02x)' % (mID, dID))
        return False
    
    #print('Flash %02x Maker ID/Device ID: %02x/%02x...OK' % (bankNum, mID, dID))
    
    # リセット
    print('Flash %02xをリセットし、1秒待機します。' % bankNum)
    resetFlash_SFM(s, bankNum)
    time.sleep(1)

    # ブランクチェック＆イレーズ
    print('Flash %02xのブランクチェック実施(20秒程度かかります)...' % bankNum, end='', flush=True)
    if blankcheck_SFM(s, bankNum) == 1:
        print('OK')
    else:
        print('ブランクではありません。')
        
        print('Flash %02xをイレーズします。' % bankNum)
        if eraseFlash_SFM(s, bankNum) == True:
            print('イレーズ成功。')
        else:
            print('Err: イレーズ失敗。')
            return False

        # リセット
        print('Flash %02xをリセットします。' % bankNum)
        resetFlash_SFM(s, bankNum)
        
        # イレーズ後のブランクチェック
        print('Flash %02xのブランクチェック実施(20秒程度かかります)...' % 
            bankNum, end='', flush=True)
        if blankcheck_SFM(s, bankNum) == 1:
            print('OK')
        else:
            print('ブランクではありません。')
            print('Err: Flash %02xのイレーズに失敗しました。' % bankNum)
            return False

    # ブランクチェックにパスしたら書き換え
    writeFlash_SFM(s, bankNum, byteArrayToWrite, startIdx)

    # リセット
    print('Flash %02xをリセットします。' % bankNum)
    resetFlash_SFM(s, bankNum)


